/*
 * "libhemca.so"
 *
 *	--- shared library for Hoshin HE1441 MCA device driver "hemca"
 *
 *  creates a command: hemca
 *
 *  its subcommands:
 *	(all but integ and demo are available only for online use)
 *
 *	online	-- open device file
 *	start	-- start data acquisition
 *	stop	-- stop  data acquisition
 *	clear	-- clear histogram
 *	read	-- read  histogram and set to vector xVec
 *	time	-- set real/live time to REAL_TIME/LIVE_TIME, respectively
 *	integ	-- summation over [INTEG(0)..INTEG(1)]
		   and set to GROSS and NET
 *	demo	-- fill histogram with silly data just for test
 *
 *  global variables
 *
 *	ONLINE
 *	ACQUIRE	-- data acquisition in progress (1) or not (0)
 *	CHAN	-- maximum channel size
 *	REAL_TIME,LIVE_TIME
 *	INTEG(0),INTEG(1)
 *	GROSS,NET
 *
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <tcl.h>
#include <blt.h>
#include <math.h>
#include "hemca.h"

static int fd, online=0;
static unsigned int hist[HEMCA_CHAN];
static Blt_Vector *xVec, *yVec;
static struct timeval t1, t2;

int Hemca_Init ( Tcl_Interp *interp ){
  int  i;
  char s[256];
  extern int hemcaCmd _ANSI_ARGS_(( ClientData clientData, 
				     Tcl_Interp *interp, 
				     int argc, char *argv[] ));
  Tcl_CreateCommand(interp, "hemca", hemcaCmd, 
		    (ClientData) NULL, (Tcl_CmdDeleteProc *)NULL );
  Tcl_SetVar (interp, "ONLINE"    , "0", TCL_GLOBAL_ONLY);
  Tcl_SetVar (interp, "ACQUIRE"   , "0", TCL_GLOBAL_ONLY);
  sprintf(s, "%d", HEMCA_CHAN);
  Tcl_SetVar (interp, "CHAN"      ,  s , TCL_GLOBAL_ONLY);
  Tcl_SetVar (interp, "REAL_TIME" , "0", TCL_GLOBAL_ONLY);
  Tcl_SetVar (interp, "LIVE_TIME" , "0", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(interp, "INTEG", "0",  "", TCL_GLOBAL_ONLY);
  Tcl_SetVar2(interp, "INTEG", "1",  "", TCL_GLOBAL_ONLY);
  Tcl_SetVar (interp, "GROSS"     ,  "", TCL_GLOBAL_ONLY);
  Tcl_SetVar (interp, "NET"       ,  "", TCL_GLOBAL_ONLY);
  if ((Blt_CreateVector(interp, "xVec", HEMCA_CHAN, &xVec) == TCL_ERROR) ||
      (Blt_CreateVector(interp, "yVec", HEMCA_CHAN, &yVec) == TCL_ERROR))
    return TCL_ERROR;
  for (i = 0; i < HEMCA_CHAN; i++) Blt_VecData(xVec)[i] = (double)i;
  memset(&(yVec->valueArr[0]), 0, HEMCA_CHAN*sizeof(double));
  if ((Blt_ResizeVector(xVec, HEMCA_CHAN) == TCL_ERROR) ||
      (Blt_ResizeVector(yVec, HEMCA_CHAN) == TCL_ERROR)) return TCL_ERROR;
  gettimeofday(&t1, NULL);
  return TCL_OK;
}

int hemcaCmd _ANSI_ARGS_(( ClientData clientData, 
			 Tcl_Interp *interp, 
			 int argc, char *argv[] )) 
{
  int  i;
  char s[256];
  if (argc < 2) {
    Tcl_SetResult(interp,"hemca error: too many argument.", TCL_STATIC);
    return TCL_ERROR; }
  if (*argv[1] == 'o' && !strcmp(argv[1], "online")) {
    if ((fd = open(DEVICE, O_RDWR)) == -1) goto offline;
    ioctl(fd, HEMCA_CLEAR, 0);
    online = 1;
    Tcl_SetVar(interp, "ONLINE"    , "1", TCL_GLOBAL_ONLY);
    return TCL_OK;
  }
  if (*argv[1] == 's' && !strcmp(argv[1], "start")) {
    if (!online) goto offline;
    Tcl_SetVar(interp, "REAL_TIME" , "0", TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "LIVE_TIME" , "0", TCL_GLOBAL_ONLY);
    gettimeofday(&t1, NULL);
    ioctl(fd, HEMCA_START, 0);
    Tcl_SetVar(interp, "ACQUIRE"   , "1", TCL_GLOBAL_ONLY);
    return TCL_OK;
  }
  if (*argv[1] == 's' && !strcmp(argv[1], "stop")) {
    if (!online) goto offline;
    ioctl(fd, HEMCA_STOP, 0);
    Tcl_SetVar(interp, "ACQUIRE"   , "0", TCL_GLOBAL_ONLY);
    return TCL_OK;
  }
  if (*argv[1] == 'c' && !strcmp(argv[1], "clear")) {
    if (!online) goto offline;
    ioctl(fd, HEMCA_CLEAR, 0);
    for (i = 0; i < HEMCA_CHAN; i++) Blt_VecData(yVec)[i] = 0.0;
    return Blt_ResizeVector(yVec, HEMCA_CHAN);
  }
  if (*argv[1] == 'r' && !strcmp(argv[1], "read")) {
    if (!online) goto offline;
    if (read(fd, hist, HEMCA_CHAN*sizeof(int)) != HEMCA_CHAN*sizeof(int)) {
      Tcl_SetResult(interp,"hemca error: failed to read data.", TCL_STATIC);
      return TCL_ERROR; }
    for (i = 0; i < HEMCA_CHAN; i++) Blt_VecData(yVec)[i] = (double)hist[i];
    return Blt_ResizeVector(yVec, HEMCA_CHAN);
  }
  if (*argv[1] == 't' && !strcmp(argv[1], "time")) {
    long ltime;
    if (!online) goto offline;
    gettimeofday(&t2, NULL);
    sprintf(s, "%d", (int)floor((t2.tv_sec -t1.tv_sec )
		      +0.000001*(t2.tv_usec-t1.tv_usec)));
    Tcl_SetVar(interp, "REAL_TIME" , s, TCL_GLOBAL_ONLY);
    ioctl(fd, HEMCA_TIME, &ltime);
    sprintf(s, "%d", (int)ltime);
    Tcl_SetVar(interp, "LIVE_TIME" , s, TCL_GLOBAL_ONLY);
    return TCL_OK;
  }
  if (*argv[1] == 'i' && !strcmp(argv[1], "integ")) {
    char  *s0, *s1;
    int    i0,  i1;
    double gross=0.0, net=0.0;
    s0 = Tcl_GetVar2(interp, "INTEG", "0", TCL_GLOBAL_ONLY);
    s1 = Tcl_GetVar2(interp, "INTEG", "1", TCL_GLOBAL_ONLY);
    if ((s0[0] == '\0') || (sscanf(s0, "%d", &i0) != 1) ||
	(s1[0] == '\0') || (sscanf(s1, "%d", &i1) != 1)) {
      Tcl_SetVar (interp, "GROSS"     ,  "", TCL_GLOBAL_ONLY);
      Tcl_SetVar (interp, "NET"       ,  "", TCL_GLOBAL_ONLY);
    } else {
      for (i = i0; i <= i1; i++) {gross +=  Blt_VecData(yVec)[i];}
      net = gross - (double)(i1-i0+1)
	*(Blt_VecData(yVec)[i1]+Blt_VecData(yVec)[i0])/2.0;
      sprintf(s, "%d", (int)floor(gross));
      Tcl_SetVar (interp, "GROSS"     ,  s, TCL_GLOBAL_ONLY);
      sprintf(s, "%d", (int)floor(net  ));
      Tcl_SetVar (interp, "NET"       ,  s, TCL_GLOBAL_ONLY);
    }
    return TCL_OK;
  }
  if (*argv[1] == 'd' && !strcmp(argv[1], "demo")) {
    for (i = 0; i < HEMCA_CHAN; i++) Blt_VecData(yVec)[i] +=
      floor(400.0*exp(-pow((double)(i-670)/40.0,2.0))
	  +1000.0*exp(-pow((double)(i-500)/40.0,2.0)));
    return Blt_ResizeVector(yVec, HEMCA_CHAN);
  }
  Tcl_SetResult(interp,"hemca error: invalid command.", TCL_STATIC);
  return TCL_ERROR;
 offline:
  Tcl_SetResult(interp,"hemca error: device is not online.", TCL_STATIC);
  return TCL_ERROR;
}
